function [X, Q, flag, k] = sdpproj(A, pol, eps, maxit)
%
% Computes the projector P, 
% such that A+ = PA is the projection of A 
% onto the SDP cone
%
% Input arguments: 
% A:     sparse band matrix
% pol:   pol==0, uses the polynomial p(x) = 3*x^2 - 2*x^3
%        pol==1, uses the polynomial p(x) = 10*x^3 - 15*x^4 + 6*x^5;
% eps:   tolerance in the stopping criteria (default: 1e-4)
% maxit: maximum number of iterations (default: 100)
%
%
% Output arguments: 
% X: sparse band matrix such that P = Q*X*Q'
% Q: cumulative orthogonal matrix from band reductions
% flag: -1 maximum number of iterations reached
%        1 (or 2) approximate projector P found
% k: number of iterations
%

if nargin<3
    eps = 1e-4; 
end

if nargin<4
   maxit = 100; 
end

n = length(A);
Id = speye(n,n);

bandthr = 0.01*n; % threshold for band reduction

% estimating the largest eigenvalue (in absolute value)
[l,u] = gersch(A);

alpha0 = abs(u);
if (abs(l)>alpha0)
    alpha0 = abs(l);
end

% shifting and scaling
B = (A + alpha0*speye(n,n))/(2*alpha0);
% the eigenvalues of B are in [0,1]

% % pre-processing (A is already positive definite)
% if (l>0)
%     X = B;
%     Q = speye(n);
%     flag = 10;
%     return
% end

% iterations for computing P
done=0;
X = B;
k=0;
Q = eye(n); % speye(n);

[C, sd] = spdiags(X);
kd = sd(length(sd));

if ( (2*kd + 1)*n > 0.01*n^2 ) % nnz(X) > 0.01*n^2
    fprintf('1st band reduction... ');
    % 0: LAPACK's DSBTRD; 1: SBR's DSBRDT
    [X, Q, info] = lapack_dsbtrd(X, kd, Q, 1);
    kd = 1;
    fprintf(' ok \n');
end

while (~done)
    k=k+1;
    
    if (k>maxit) % Maximum number of iterations reached
       done=1;
       flag=-1;
       break
    end
    
    X2 = X*X;
    
    if (norm(X - X2, 'fro')*alpha0 < eps )
        done=1;
        flag=1;
        break
    end
    
    X3 = X*X2;
    
    X0=X;
    
    if pol==1 
        % P(X) = 10*X^3 - 15*X^4 + 6*X^5;
        X = X3*(10*Id - 15*X + 6*X2);
        
    else
        % P(X) = 3*X^2 - 2*X^3
        X = 3*X2 - 2*X3;

    end
    
    % symmetrization
    X = (X + X')/2;
    
    if (norm(X - X0, 'fro')*alpha0 < eps )
        done=1;
        flag=2;
        break
    end
    
    % band reduction    
    [C, sd] = spdiags(X);
    kd = sd(length(sd));
    
    if ( kd > bandthr ) % (kd*kd > n) 
        fprintf('band reduction (it: %d kd: %d) ... ', k, kd);
       
       % SBR's DSBRDT routine for band reduction
       [X, Q, info] = lapack_dsbtrd(X,kd,Q,1);
       
       fprintf('ok! \n');
       kd = 1;
    end

end

kd
k
flag

end

function [l,u] = gersch(A)
% estimating the largest eigenvalue (in absolute value)

%     % estimate eigenvalue range from Gerschgorin's circles
%     n = length(A);
%     l = 1e+20;
%     u = 1e-20;
% 
%     for i=1:n
%        r = sum(abs(A(i,1:n))) - abs(A(i,i));
%        c = A(i,i);
% 
%        if (c-r < l) 
%            l = c - r;
%        end
% 
%        if (c+r > u)
%            u = c+r;
%        end
% 
%     end

    l = normest(A, 1e-4);
    u = l;

end